import math
import numpy as np

def get_system_planets(data, do_sort = True):
    """Returns a dictionary where key = hostname and value = list of planets from data which is a list of planet data"""
    hosts = {}
    for planet_id in range(len(data)):
        planet = data[planet_id]
        hostname = planet['hostname']
        if not hostname in hosts:
            hosts[hostname] = []
        hosts[hostname].append(planet_id)
    if do_sort:
        for hostname, planets in hosts.items():
            all_planets_valid = True
            for i in planets:
                if math.isnan(data[i]['pl_orbsmax']):
                    all_planets_valid = False
                    break
            
            if all_planets_valid:
                hosts[hostname] = sorted(planets, key=lambda i: data[i]['pl_orbsmax'] or 0)
            else:
                hosts[hostname] = planets
            all_planets_valid = True
    return hosts

def get_system_data(data):
    sys_planets = get_system_planets(data)

    for name, list in sys_planets.items():
        sys_planets[name] = [data[id] for id in list]

    return sys_planets

def get_system_pairs(data, do_sort = True):
    """
        Get pairs of planets.

        Args:
            'do_sort': Should the planets be sorted by distance from host star
    """
    system_planets = get_system_planets(data, do_sort)
    pairs = []
    for planets in system_planets.values():
        if len(planets) > 1:
            for i in range(len(planets)):
                if (i+1)%2==0:
                    pairs.append((data[planets[i-1]], data[planets[i]]))
    return pairs

def is_num_clean(n):
    return not (math.isinf(n) or math.isnan(n))

def get_pair_data(pairs, label, clean=True, **kwargs):
    modifier = kwargs.get("modifier")
    pair_data = []
    pair_ids = []

    for i in range(len(pairs)):
        pair = pairs[i]
        valA, valB = pair[0].get(label), pair[1].get(label)
        if valA and valB:
            if modifier:
                valA, valB = modifier(valA), modifier(valB)
            if clean:
                if is_num_clean(valA) and is_num_clean(valB): #valA != 0 and valB != 0
                    pair_data.append([valA, valB])
                    pair_ids.append(i)
                #else:
                #should it append something here to keep indices consistent?
            else:
                pair_data.append([valA, valB])
                pair_ids.append(i)
        #else:
        #should it append here too to keep indices consistent?
    return pair_data, pair_ids

def get_is_outlier(array, max_deviations = 2):
    mean = np.mean(array)
    std = np.std(array)
    # dist_from_mean = abs(array - mean)
    is_outlier = lambda x: abs(x-mean) > max_deviations * std
    return is_outlier

def remove_outliers_from_both(a, b, md):
    is_outlier = get_is_outlier(a, md)
    removes = 0
    for i in range(len(a)):
        i = i-removes
        if is_outlier(a[i]):
            removes += 1
            a.pop(i)
            b.pop(i)

# def get_datas(data, label, removeInvalid = True):
#     datas = []
#     for pl in data:
#         if removeInvalid:
#             val = pl[label]
#             if is_num_clean(val):
#                 datas.append(val)
#         else:
#             datas.append(pl[label])
#     return datas

def remove_any_nan(data, labelA, labelB):
    listA = []
    listB = []

    for pl in data:
        valA = pl[labelA]
        valB = pl[labelB]
        if is_num_clean(valA) and is_num_clean(valB):
            listA.append(valA)
            listB.append(valB)

    return listA, listB

def remove_nan_lists(listA, listB):
    removed = 0
    for i in range(len(listA)):
        i -= removed
        valA = listA[i]
        valB = listB[i]
        if not is_num_clean(valA) or not is_num_clean(valB):
            removed += 1
            listA.pop(i)
            listB.pop(i)



import os
from matplotlib import pyplot as plt
from matplotlib.ticker import ScalarFormatter
import numpy as np
from ..util import get_readable, get_pair_data, remove_outliers_from_both, remove_any_nan, get_system_data, is_num_clean

def save_plt(plt, file_path, file_name):
    full_path = file_path + file_name
    print("Saving: '" + full_path)
    try:
        plt.savefig(full_path)
    except:
        os.mkdir(file_path)
        plt.savefig(full_path)

def plot_pair_ratio(pair_data, plotLabel, **kwargs):
    """
    Creates a graph comparing the ratio of a specified data in pairs
    of planets to the first planet in the pair.

    Arguments:
        'use_log10': Base values in log10 (recommended for large data)
        'max_deviations': Removes outliers which deviate from the mean
                            by the standard deviation * this value.
        'file_path': File path to save graphs to.
        'file_name': Custom name for the file.
    """

    plt.clf()

    # setup arguments
    defaultKwargs = {
        'use_log10': True,
        'max_deviations': None,
        'file_path': "./graphs/",
        'file_name': f"Graph_of_{plotLabel}_Pair_Ratios.png",
    }
    kwargs = { **defaultKwargs, **kwargs }
    max_deviations = kwargs.get('max_deviations') or None

    # define axes
    extracted, _ = get_pair_data(pair_data, plotLabel)
    # this could be optimized and turned into one for loop but i like the
    # clean syntax of this solution
    x = [vals[0]/vals[1] for vals in extracted]
    y = [vals[0] for vals in extracted]

    # remove outliers
    if max_deviations:
        remove_outliers_from_both(x, y, max_deviations)

    # create the plot
    plt.scatter(x, y, color="red")

    if kwargs.get('use_log10'):
        axes = plt.gca()
        axes.set_yscale('log')
        axes.set_xscale('log')

    title = get_readable(plotLabel)
    if max_deviations:
        title = title + " (|x-x̄| < " + str(max_deviations) + "σ)"
    plt.title(title)
    plt.xlabel("Ratio of A/B")
    plt.ylabel("A of Pair")
    save_plt(plt, kwargs.get("file_path"), kwargs.get("file_name"))

def plot_dual(data, x_label, y_label, **kwargs):
    """
    Plot chart with two different categories as x and y
    """

    plt.clf()

    # setup arguments
    defaultKwargs = {
        'use_log10': True,
        'file_path': "./graphs/",
        'file_name': f"Graph_of_{x_label}_and_{y_label}.png",
    }
    kwargs = { **defaultKwargs, **kwargs }

    # get axes for
    x, y = remove_any_nan(data, x_label, y_label)

    plt.scatter(x, y, color="red")

    if kwargs.get('use_log10'):
        axes = plt.gca()
        axes.set_yscale('log')
        axes.set_xscale('log')

    x_readable = get_readable(x_label)
    y_readable = get_readable(y_label)
    plt.title(x_readable+" and "+y_readable)
    plt.xlabel(x_readable)
    plt.ylabel(y_readable)
    save_plt(plt, kwargs.get("file_path"), kwargs.get("file_name"))

def plot_system_average(data, main_label, avg_label, **kwargs):
    """
        Compare average of system planets to a value of the first planet in the system.
    """

    plt.clf()

    defaultKwargs = {
        'use_log10_main': False,
        'use_log10_avg': True,
        'file_path': "./graphs/",
        'file_name': f"Graph_of_avg_{main_label}_and_{avg_label}.png",
    }
    kwargs = { **defaultKwargs, **kwargs }


    sys_data = get_system_data(data)
    avgs = {}
    for host, pls in sys_data.items():
        avg = 0
        for pl in pls:
            avg += pl[avg_label]
        avg/=len(pls)
        avgs[host] = avg
    
    x = [sys[0][main_label] for _, sys in sys_data.items()]
    y = avgs.values()
    
    plt.scatter(x, y, color="red")

    axes = plt.gca()
    if kwargs.get('use_log10_main'):
        axes.set_xscale('log')
    if kwargs.get('use_log10_avg'):
        axes.set_yscale('log')
    x_readable = get_readable(main_label)
    y_readable = "Avg of " + get_readable(avg_label)
    plt.title(x_readable+" and "+y_readable)
    plt.xlabel(x_readable)
    plt.ylabel(y_readable)
    save_plt(plt, kwargs.get("file_path"), kwargs.get("file_name"))

def plot_ratio_to_greatest(data, label, **kwargs):
    """
        Plots ratio 
    """
    defaultKwargs = {
        'use_log10': True,
        'file_path': "./graphs/",
        'file_name': f"Graph_of_greatest_{label}.png",
    }
    kwargs = { **defaultKwargs, **kwargs }

    x = []
    y = []

    sys_data = get_system_data(data)
    for host, pls in sys_data.items():
        if len(pls) < 2:
            continue
        total = 0
        greatest = -1e15
        for pl in pls:
            val = pl[label]
            total += val
            greatest = max(greatest, val)
        ratio = greatest/total
        x.append(ratio*100)
        y.append(greatest)
    
    # remove_nan_lists(x, y)
    
    plt.scatter(x, y, color="red")

    if kwargs.get('use_log10'):
        axes = plt.gca()
        axes.set_yscale('log')
    x_readable = "% of greatest to total"
    y_readable = "Greatest " + get_readable(label)
    plt.title(x_readable+" and "+y_readable)
    plt.xlabel(x_readable)
    plt.ylabel(y_readable)
    save_plt(plt, kwargs.get("file_path"), kwargs.get("file_name"))

def plot_pair_comparison(pair_data, label, categories, **kwargs):
    """
    Creates a graph which categorizes pairs into specified categories.

    Arguments:
        'file_path': File path to save graphs to.
        'file_name': Custom name for the file.
    """

    plt.clf()

    # setup arguments
    defaultKwargs = {
        'file_path': "./graphs/",
        'file_name': f"Graph_of_{label}_Pair_Bar_Comparison.png",
    }
    kwargs = { **defaultKwargs, **kwargs }

    # define axes
    extracted, pair_ids = get_pair_data(pair_data, label)

    positions = [*range(0, len(categories)+1)]
    values = {}
    for pair in extracted:
        memberOf = None
        for name, category in categories.items():
            isMember = category(*pair)
            if isMember:
                memberOf = name
                break
        if memberOf:
            if not values.get(memberOf):
                values[memberOf] = 0
            values[memberOf] += 1
        else:
            print("Other:", pair_data[pair_ids[extracted.index(pair)]][0]["pl_name"], pair_data[pair_ids[extracted.index(pair)]][1]["pl_name"])
            if not values.get("Other"):
                values["Other"] = 0
            values["Other"] += 1
    
    value_list = []
    category_names = list(categories.keys())
    if values.get("Other"):
        category_names.append("Other")
    for name in category_names:
        value_list.append(values[name])

    print(positions, value_list)

    # create the plot
    bar = plt.bar(positions, value_list, color="red", width=0.75)
    
    ax = plt.gca()
    labels = [""]
    for bar_label in category_names:
        labels.append("")
        labels.append(bar_label)
    ax.set_xticklabels(labels)
    total = len(extracted)
    labels = []
    for val in value_list:
        labels.append(str(round(val/total * 1000)/10) + "%")
    ax.bar_label(bar, labels=labels)

    plt.title(get_readable(label))
    # plt.xlabel("Category")
    plt.ylabel("# of Planets")
    save_plt(plt, kwargs.get("file_path"), kwargs.get("file_name"))

def plot_pl_categories(data, label, categories, **kwargs):
    """
    Creates a graph which categorizes planets into specified categories.

    Arguments:
        'file_path': File path to save graphs to.
        'file_name': Custom name for the file.
    """

    plt.clf()

    # setup arguments
    defaultKwargs = {
        'file_path': "./graphs/",
        'file_name': f"Bar_Categories_of_{label}.png",
    }
    kwargs = { **defaultKwargs, **kwargs }

    # define axes

    positions = [*range(0, len(categories)+1)]
    values = {}
    for pl in data:
        val = pl[label]
        # if not is_num_clean(val):
        #     continue
        memberOf = None
        for name, category in categories.items():
            isMember = category(val)
            if isMember:
                memberOf = name
                break
        if memberOf:
            if not values.get(memberOf):
                values[memberOf] = 0
            values[memberOf] += 1
        else:
            print("Other:", pl["pl_name"])
            if not values.get("Other"):
                values["Other"] = 0
            values["Other"] += 1
    
    value_list = []
    category_names = list(categories.keys())
    if values.get("Other"):
        category_names.append("Other")
    for name in category_names:
        value_list.append(values[name])

    print(positions, value_list)

    # create the plot
    bar = plt.bar(positions, value_list, color="red", width=0.75)
    
    ax = plt.gca()
    labels = [""]
    for bar_label in category_names:
        labels.append("")
        labels.append(bar_label)
    ax.set_xticklabels(labels)
    total = len(data)
    labels = []
    for val in value_list:
        labels.append(str(round(val/total * 1000)/10) + "%")
    ax.bar_label(bar, labels=labels)

    plt.title(get_readable(label))
    # plt.xlabel("Category")
    plt.ylabel("# of Planets")
    save_plt(plt, kwargs.get("file_path"), kwargs.get("file_name"))


"""Queries the NASA Exoplanet Archive's TAP API."""

from numpy import NaN
import pyvo as vo
from ..util import get_label_list

_CLIENT_URL = "https://exoplanetarchive.ipac.caltech.edu/TAP"
# _DEFAULT_COLUMNS = ["pl_name",
#     "hostname",
#     "disc_year",
#     "discoverymethod",
#     "pl_orbper",
#     "pl_orbsmax",
#     "pl_orbeccen",
#     "pl_rade",
#     "pl_masse",
#     "pl_bmasse",
#     "st_met",
#     "st_teff",
#     "st_metratio",
#     "st_mass"]
_DEFAULT_COLUMNS = get_label_list()
_DEFAULT_DB = "ps"

# Database to query. View databases at https://exoplanetarchive.ipac.caltech.edu/docs/TAP/usingTAP.html
_QUERY_FORMAT = lambda columns, db: "SELECT " + ",".join(columns) + " FROM " + db# + " WHERE " # need to add constraint

def query(**kwargs):
    """
    Queries NASA Exoplanet TAP interface to retrieve data.
    Returns a list of dictionaries in format {column: data}
    """
    defaultKwargs = {
        "query_columns": _DEFAULT_COLUMNS,
        "query_db": _DEFAULT_DB,
        "do_not_process": False
    }
    kwargs = { **defaultKwargs, **kwargs }
    columns = kwargs["query_columns"]
    db = kwargs["query_db"]
    do_not_process = kwargs["do_not_process"]

    print("RETRIEVING DATA")

    service = vo.dal.TAPService(_CLIENT_URL)
    result = service.search(_QUERY_FORMAT(columns, db))

    print("DATA RETRIEVED")
    if do_not_process:
        return result
    else:
        print("PROCESSING DATA")

        # restructure data into a list of planets
        pl_list = []
        used_names = []
        for row in result:
            if row["pl_name"] in used_names:
                dict = pl_list[used_names.index(row["pl_name"])]
                for columnLabel in columns:
                    val = row[columnLabel]
                    dict[columnLabel] = val or NaN
            else:
                dict = {}
                for columnLabel in columns:
                    val = row[columnLabel]
                    dict[columnLabel] = val or NaN
                pl_list.append(dict)
                used_names.append(row["pl_name"])

        print("PROCESSED DATA")

        print("Exoplanet Entries:", len(pl_list))
        return pl_list


# readable names for data labels
_READABLE_LABELS = {'pl_name': 'Planet Name',
    'hostname': 'Host Name',
    'default_flag': 'Default Parameter Set',
    'sy_snum': 'Number of Stars',
    'sy_pnum': 'Number of Planets',
    'discoverymethod': 'Discovery Method',
    'disc_year': 'Discovery Year',
    'disc_facility': 'Discovery Facility',
    'soltype': 'Solution Type',
    'pl_controv_flag': 'Controversial Flag',
    'pl_refname': 'Planetary Parameter Reference',
    'pl_orbper': 'Orbital Period [days]',
    'pl_orbsmax': 'Orbit Semi-Major Axis [au]',
    'pl_rade': 'Planet Radius [Earth Radius]',
    'pl_radj': 'Planet Radius [Jupiter Radius]',
    'pl_masse': 'Planet Mass [Earth Mass]',
    'pl_massj': 'Planet Mass [Jupiter Mass]',
    'pl_bmasse': 'Planet Mass or Mass*sin(i) [Earth Mass]',
    'pl_bmassj': 'Planet Mass or Mass*sin(i) [Jupiter Mass]',
    'pl_dens': 'Planet Density [g/cm**3]',
    'pl_orbeccen': 'Eccentricity',
    'pl_insol': 'Insolation Flux [Earth Flux]',
    'pl_eqt': 'Equilibrium Temperature [K]',
    'ttv_flag': 'Data show Transit Timing Variations',
    'st_refname': 'Stellar Parameter Reference',
    'st_spectype': 'Spectral Type',
    'st_teff': 'Stellar Effective Temperature [K]',
    'st_rad': 'Stellar Radius [Solar Radius]',
    'st_mass': 'Stellar Mass [Solar mass]',
    'st_met': 'Stellar Metallicity [dex]',
    'st_metratio': 'Stellar Metallicity Ratio',
    'st_logg': 'Stellar Surface Gravity [log10(cm/s**2)]',
    'sy_refname': 'System Parameter Reference',
    'rastr': 'RA [sexagesimal]',
    'ra': 'RA [deg]',
    'decstr': 'Dec [sexagesimal]',
    'dec': 'Dec [deg]',
    'sy_dist': 'Distance [pc]',
    'sy_vmag': 'V (Johnson) Magnitude',
    'sy_kmag': 'Ks (2MASS) Magnitude',
    'sy_gaiamag': 'Gaia Magnitude',
    'rowupdate': 'Date of Last Update',
    'pl_pubdate': 'Planetary Parameter Reference Publication Date',
    'releasedate': 'Release Date'}

def get_label_list():
    list = []
    for label, _ in _READABLE_LABELS.items():
            list.append(label)
    return list

def get_readable(label):
    return _READABLE_LABELS[label]